using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using MVR;
using MVR.FileManagementSecure;

// VAMSWavRecorder v1.0
//
// Bake a VAM session into a wav file

namespace VAMStory
{
    public class VAMSWavRecorder : MVRScript
    {	
		// Static string
		private static string PLUGIN_PATH;
		private static string EXCEPTION_PREFIX = "VAMSWavRecorder - Exception caught: ";
		private static string SAVE_FOLDER = "Saves/PluginData/vamstory/wavrecorder";

		private AudioSource AS;

		private int timeCaptured = 0;
		private bool VAMSWRReady = false;

		private AudioListener mainListener;
		private AudioListenerRecorder ALR;
		private AudioClip RecordedAudioClip;

		public JSONStorableBool ReverbEnabled;
		public JSONStorableStringChooser ReverbPreset;
		protected List<string> presetsChoices;
		protected List<AudioReverbPreset> presetsList;
		private AudioReverbFilter ReverbFilterComp;
		
		public JSONStorableAction A_StartRecording;
		public JSONStorableAction A_StopRecording;
		public JSONStorableAction A_PreviewRecording;
		public JSONStorableAction A_StopPreview;
		public JSONStorableAction A_SaveRecording;
		
		public override void Init()
        {			 
            try
            {
                if (containingAtom.type != "AudioSource")
                {
                    SuperController.LogError("Please add VAMSWavRecorder on an AudioSource Atom");
                    return;
                }
                
	            PLUGIN_PATH = Utils.GetPluginPath(this);
	            
	            FileManagerSecure.CreateDirectory(SAVE_FOLDER);
	            
				// Reverb Choises
				presetsChoices = new List<string>();
				presetsChoices.Add("Generic");
				presetsChoices.Add("PaddedCell");
				presetsChoices.Add("Room");
				presetsChoices.Add("Bathroom");
				presetsChoices.Add("Livingroom");
				presetsChoices.Add("Stoneroom");
				presetsChoices.Add("Auditorium");
				presetsChoices.Add("Concerthall");
				presetsChoices.Add("Cave");
				presetsChoices.Add("Arena");
				presetsChoices.Add("Hangar");
				presetsChoices.Add("CarpetedHallway");
				presetsChoices.Add("Hallway");
				presetsChoices.Add("StoneCorridor");
				presetsChoices.Add("Alley");
				presetsChoices.Add("Forest");
				presetsChoices.Add("City");
				presetsChoices.Add("Mountains");
				presetsChoices.Add("Quarry");
				presetsChoices.Add("Plain");
				presetsChoices.Add("ParkingLot");
				presetsChoices.Add("SewerPipe");
				presetsChoices.Add("Underwater");
				presetsChoices.Add("Drugged");
				presetsChoices.Add("Dizzy");
				presetsChoices.Add("Psychotic");
				
				presetsList = new List<AudioReverbPreset>();
				presetsList.Add(AudioReverbPreset.Generic);
				presetsList.Add(AudioReverbPreset.PaddedCell);
				presetsList.Add(AudioReverbPreset.Room);
				presetsList.Add(AudioReverbPreset.Bathroom);
				presetsList.Add(AudioReverbPreset.Livingroom);
				presetsList.Add(AudioReverbPreset.Stoneroom);
				presetsList.Add(AudioReverbPreset.Auditorium);
				presetsList.Add(AudioReverbPreset.Concerthall);
				presetsList.Add(AudioReverbPreset.Cave);
				presetsList.Add(AudioReverbPreset.Arena);
				presetsList.Add(AudioReverbPreset.Hangar);
				presetsList.Add(AudioReverbPreset.CarpetedHallway);
				presetsList.Add(AudioReverbPreset.Hallway);
				presetsList.Add(AudioReverbPreset.StoneCorridor);
				presetsList.Add(AudioReverbPreset.Alley);
				presetsList.Add(AudioReverbPreset.Forest);
				presetsList.Add(AudioReverbPreset.City);
				presetsList.Add(AudioReverbPreset.Mountains);
				presetsList.Add(AudioReverbPreset.Quarry);
				presetsList.Add(AudioReverbPreset.Plain);
				presetsList.Add(AudioReverbPreset.ParkingLot);
				presetsList.Add(AudioReverbPreset.SewerPipe);
				presetsList.Add(AudioReverbPreset.Underwater);
				presetsList.Add(AudioReverbPreset.Drugged);
				presetsList.Add(AudioReverbPreset.Dizzy);
				presetsList.Add(AudioReverbPreset.Psychotic);

	            Utils.CreateStaticDescriptionText(this,"VAMSWavRecorderTitle","<color=#000><size=35><b>VAMSWavRecorder</b></size></color>",false,40,TextAnchor.MiddleLeft);
				Utils.CreateStaticDescriptionText(this,"Right UI Main Title","<color=#000><size=28>A tool to record the audio from a scene</size></color>",false,135,TextAnchor.MiddleLeft);
				
				ReverbEnabled = new JSONStorableBool("Reverb enabled", false, ReverbEnabledCallback);
				ReverbEnabled.storeType = JSONStorableParam.StoreType.Full;
				CreateToggle(ReverbEnabled, false);
				
				ReverbPreset = new JSONStorableStringChooser("Reverb preset", presetsChoices, "Generic", "Reverb preset", ReverbPresetCallback);
				UIDynamicPopup FAdp = CreateScrollablePopup(ReverbPreset, false);
				FAdp.labelWidth = 230f;

				AS = Utils.GetChildNamed(containingAtom.reParentObject, "AudioSource").GetComponent<AudioSource>();
				mainListener = SuperController.singleton.CurrentAudioListener;
				
				ReverbFilterComp = mainListener.gameObject.GetComponent<AudioReverbFilter>();
				if (ReverbFilterComp == null) ReverbFilterComp = mainListener.gameObject.AddComponent<AudioReverbFilter>();
				ReverbFilterComp.reverbPreset = AudioReverbPreset.Generic;
				ReverbFilterComp.enabled = false;
				
				ALR = mainListener.gameObject.AddComponent<AudioListenerRecorder>();
				
                // ******* CREATING MAIN TRIGGERS/ACTIONS ********
                // They are here to be displayed on TOP of every other JSONStorables
                JSONStorableAction fakeFuncUseBelow = new JSONStorableAction("- - - - Use these functions below ↓ - - - - -", () => {});
                RegisterAction(fakeFuncUseBelow);
                
                A_StartRecording = new JSONStorableAction("Start Recording", () => { StartRecording(); });
                RegisterAction(A_StartRecording);
                
                A_StopRecording = new JSONStorableAction("Stop Recording", () => { StopRecording(); });
                RegisterAction(A_StopRecording);
                
                A_PreviewRecording = new JSONStorableAction("Preview Recording", () => { PreviewWav(); });
                RegisterAction(A_PreviewRecording);
                
                A_StopPreview = new JSONStorableAction("Stop Preview", () => { PreviewWavStop(); });
                RegisterAction(A_StopPreview);
                
                A_SaveRecording = new JSONStorableAction("Save Recording", () => { SaveWav(); });
                RegisterAction(A_SaveRecording);

                RegisterBool(ReverbEnabled);
                RegisterStringChooser(ReverbPreset);
                
                VAMSWRReady = true;
            }
            catch(Exception e)
            {
                SuperController.LogError(EXCEPTION_PREFIX + e);
            }
		}
		
		void Start()
        {
			try
            {

			}
            catch(Exception e)
            {
                SuperController.LogError(EXCEPTION_PREFIX + e);
            }
        }
		
		void Update() {

		}

		// **************************
		// Functions
		// **************************
		private void StartRecording()
		{
			if (!VAMSWRReady) return;
			ALR.StartRecording();
		}
		
		private void StopRecording() {
			if (!VAMSWRReady) return;
			ALR.StopRecording();
			RecordedAudioClip = ALR.GetRecordedAudioClip();
		}
		
		private void PreviewWav() {
			if (!VAMSWRReady) return;
			AS.clip = RecordedAudioClip;
            AS.Play();
		}
		
		private void PreviewWavStop() {
			if (!VAMSWRReady) return;
			AS.Stop();
		}
		
		private void SaveWav() {
			if (!VAMSWRReady) return;
			
			if (!ALR.recordingDone)
			{
				SuperController.singleton.Alert("<b><color=green>VAMSWavRecorder</color></b>\nNo recording has been made or it is currently recording.", () => {});
				return;
			}
			
			float[] samples = new float[RecordedAudioClip.samples];
			RecordedAudioClip.GetData(samples, 0);
			
			byte[] data = ConvertSamplesToByteArray(samples);
			byte[] header = CreateWaveHeader(RecordedAudioClip);

			byte[] wavFile = new byte[header.Length + data.Length];
			header.CopyTo(wavFile, 0);
			data.CopyTo(wavFile, header.Length);
			
			FileManagerSecure.WriteAllBytes(SAVE_FOLDER + "/wavrecorder-" + DateTime.Now.ToString("yyyyMMdd-HHmmss") + ".wav", wavFile);
			
			SuperController.singleton.Alert("<b><color=green>VAMSWavRecorder</color></b>\nWav file saved in:\n " + SAVE_FOLDER + "/wavrecorder-" + DateTime.Now.ToString("yyyyMMdd-HHmmss") + ".wav", () => {});
		}
		
		private byte[] ConvertSamplesToByteArray(float[] samples)
		{
			byte[] byteArray = new byte[samples.Length * 2];

			for (int i = 0; i < samples.Length; i++)
			{
				short value = (short)(samples[i] * 32767f);
				byteArray[i * 2] = (byte)(value & 0xFF);
				byteArray[(i * 2) + 1] = (byte)(value >> 8);
			}

			return byteArray;
		}
		
		private byte[] CreateWaveHeader(AudioClip clip) {
			try
			{
			    int headerSize = 44;
			    int dataSize = clip.samples * clip.channels * 2;
			    int fileSize = headerSize + dataSize - 8;

			    // RIFF header
			    byte[] riffHeader = System.Text.Encoding.UTF8.GetBytes("RIFF");
			    byte[] fileSizeBytes = BitConverter.GetBytes(fileSize);
			    byte[] waveHeader = System.Text.Encoding.UTF8.GetBytes("WAVE");

			    // format chunk
			    byte[] formatChunkId = System.Text.Encoding.UTF8.GetBytes("fmt ");
			    byte[] formatChunkSizeBytes = BitConverter.GetBytes(16);
			    byte[] audioFormat = BitConverter.GetBytes(1);
			    byte[] numChannels = BitConverter.GetBytes(clip.channels);
			    byte[] sampleRateBytes = BitConverter.GetBytes(clip.frequency);
			    byte[] byteRate = BitConverter.GetBytes(clip.frequency * clip.channels * 2);
			    byte[] blockAlign = BitConverter.GetBytes(clip.channels * 2);
			    byte[] bitsPerSampleBytes = BitConverter.GetBytes(16);

			    // data chunk
			    byte[] dataChunkId = System.Text.Encoding.UTF8.GetBytes("data");
			    byte[] dataSizeBytes = BitConverter.GetBytes(dataSize);

			    // header
			    byte[] header = new byte[headerSize];
			    System.Array.Copy(riffHeader, 0, header, 0, 4);
			    System.Array.Copy(fileSizeBytes, 0, header, 4, 4);
			    System.Array.Copy(waveHeader, 0, header, 8, 4);
			    System.Array.Copy(formatChunkId, 0, header, 12, 4);
			    System.Array.Copy(formatChunkSizeBytes, 0, header, 16, 4);
			    System.Array.Copy(audioFormat, 0, header, 20, 2);
			    System.Array.Copy(numChannels, 0, header, 22, 2);
			    System.Array.Copy(sampleRateBytes, 0, header, 24, 4);
			    System.Array.Copy(byteRate, 0, header, 28, 4);
			    System.Array.Copy(blockAlign, 0, header, 32, 2);
			    System.Array.Copy(bitsPerSampleBytes, 0, header, 34, 2);
			    System.Array.Copy(dataChunkId, 0, header, 36, 4);
			    System.Array.Copy(dataSizeBytes, 0, header, 40, 4);

			    return header;
			}
			catch(Exception e)
			{
				SuperController.LogError(EXCEPTION_PREFIX + e);
				return null;
			}
		}

		// **************************
		// Callbacks & similar
		// **************************
		protected void ReverbEnabledCallback(bool val) {
			if( VAMSWRReady ) {
				ReverbFilterComp.enabled = val;
			}
		}
		
		protected void ReverbPresetCallback(string choice) {
			if( VAMSWRReady ) {
				ReverbFilterComp.reverbPreset = presetsList[ presetsChoices.IndexOf(choice) ];
			}
		}
		
		
		
		// **************************
		// Time to cleanup !
		// **************************
		void OnDestroy()
		{
			Destroy(mainListener.gameObject.GetComponent<AudioListenerRecorder>());
			Destroy(mainListener.gameObject.GetComponent<AudioReverbFilter>());
		}

		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
    }

    public class AudioListenerRecorder : MonoBehaviour
    {
	    public bool isRecording = false;
	    public bool recordingDone = false;
	    
	    private float[] audioData;
		
	    private List<float> outputValues;
	    private float[] outputValuesFloat;
	    private float elapsedTime = 0f;
	    
	    public AudioClip audioClip;

	    public void StartRecording()
	    {
			if( outputValues != null ) outputValues.Clear();
			
		    outputValues = new List<float>();
		    elapsedTime = 0f;

		    isRecording = true;
		    recordingDone = false;
	    }

	    public void StopRecording()
	    {
		    if (!isRecording) return;
		    
		    isRecording = false;
		    
		    outputValuesFloat = new float[outputValues.Count];
		    int j = 0;
		    foreach (float thisFloat in outputValues)
		    {
			    outputValuesFloat[j] = thisFloat;
			    j++;
		    }
		    recordingDone = true;
	    }

	    public AudioClip GetRecordedAudioClip()
	    {
			if( !recordingDone ) return null;
			
		    int samplesLength = outputValuesFloat.Length;
				
		    audioClip = AudioClip.Create("thisAudio", samplesLength, 2, AudioSettings.outputSampleRate, false);
		    audioClip.SetData(outputValuesFloat, 0);

		    return audioClip;
	    }
	    
	    public void OnAudioFilterRead(float[] audioData, int channels)
	    {
		    if (isRecording)
		    {
			    elapsedTime += Time.deltaTime;
			    for (int i = 0; i < audioData.Length; i++)
			    {
				    outputValues.Add(audioData[i]);
			    }
		    }
	    }
	    
	    // **************************
	    // Local Tools
	    // **************************
	    private void logDebug( string debugText ) {
		    SuperController.LogMessage( debugText );
	    }
    }
}